package java.util.concurrent.locks;

import java.util.concurrent.TimeUnit;
import java.util.Date;
import java.util.ArrayList;
import java.util.Collection;
import sun.misc.Unsafe;
import java.io.Serializable;

public abstract class AbstractQueuedSynchronizer extends AbstractOwnableSynchronizer implements Serializable
{
    private static final long serialVersionUID = 7373984972572414691L;
    private transient volatile AbstractQueuedSynchronizer.Node head;
    private transient volatile AbstractQueuedSynchronizer.Node tail;
    private volatile int state;
    static final long spinForTimeoutThreshold = 1000L;
    private static final Unsafe unsafe;
    private static final long stateOffset;
    private static final long headOffset;
    private static final long tailOffset;
    private static final long waitStatusOffset;
    private static final long nextOffset;
    
    protected AbstractQueuedSynchronizer() {
    }
    
    protected final int getState() {
        return this.state;
    }
    
    protected final void setState(int state) {
        this.state = state;
    }
    
    protected final boolean compareAndSetState(int expected, int x) {
        return AbstractQueuedSynchronizer.unsafe.compareAndSwapInt(this, AbstractQueuedSynchronizer.stateOffset, expected, x);
    }
    
    private AbstractQueuedSynchronizer.Node enq(AbstractQueuedSynchronizer.Node next) {
        AbstractQueuedSynchronizer.Node tail;
        while (true) {
            tail = this.tail;
            if (tail == null) {
                if (this.compareAndSetHead(new AbstractQueuedSynchronizer.Node())) {
                    this.tail = this.head;
                }
                else {
                    continue;
                }
            }
            else {
                next.prev = tail;
                if (this.compareAndSetTail(tail, next)) {
                    break;
                }
                continue;
            }
        }
        tail.next = next;
        return tail;
    }
    
    private AbstractQueuedSynchronizer.Node addWaiter(AbstractQueuedSynchronizer.Node node) {
        AbstractQueuedSynchronizer.Node next = new AbstractQueuedSynchronizer.Node(Thread.currentThread(), node);
        AbstractQueuedSynchronizer.Node tail = this.tail;
        if (tail != null) {
            next.prev = tail;
            if (this.compareAndSetTail(tail, next)) {
                return tail.next = next;
            }
        }
        this.enq(next);
        return next;
    }
    
    private void setHead(AbstractQueuedSynchronizer.Node head) {
        this.head = head;
        head.thread = null;
        head.prev = null;
    }
    
    private void unparkSuccessor(AbstractQueuedSynchronizer.Node node) {
        int waitStatus = node.waitStatus;
        if (waitStatus < 0) {
            compareAndSetWaitStatus(node, waitStatus, 0);
        }
        AbstractQueuedSynchronizer.Node next = node.next;
        if (next == null || next.waitStatus > 0) {
            next = null;
            for (AbstractQueuedSynchronizer.Node node2 = this.tail; node2 != null && node2 != node; node2 = node2.prev) {
                if (node2.waitStatus <= 0) {
                    next = node2;
                }
            }
        }
        if (next != null) {
            LockSupport.unpark(next.thread);
        }
    }
    
    private void doReleaseShared() {
        while (true) {
            AbstractQueuedSynchronizer.Node head = this.head;
            if (head != null && head != this.tail) {
                int waitStatus = head.waitStatus;
                if (waitStatus == -1) {
                    if (!compareAndSetWaitStatus(head, -1, 0)) {
                        continue;
                    }
                    this.unparkSuccessor(head);
                }
                else if (waitStatus == 0 && !compareAndSetWaitStatus(head, 0, -3)) {
                    continue;
                }
            }
            if (head == this.head) {
                break;
            }
            continue;
        }
    }
    
    private void setHeadAndPropagate(AbstractQueuedSynchronizer.Node head, int n) {
        AbstractQueuedSynchronizer.Node head2 = this.head;
        this.setHead(head);
        AbstractQueuedSynchronizer.Node head3;
        if (n > 0 || head2 == null || head2.waitStatus < 0 || (head3 = this.head) == null || head3.waitStatus < 0) {
            AbstractQueuedSynchronizer.Node next = head.next;
            if (next == null || next.isShared()) {
                this.doReleaseShared();
            }
        }
    }
    
    private void cancelAcquire(AbstractQueuedSynchronizer.Node next) {
        if (next == null) {
            return;
        }
        next.thread = null;
        AbstractQueuedSynchronizer.Node prev;
        for (prev = next.prev; prev.waitStatus > 0; prev = (next.prev = prev.prev)) {}
        AbstractQueuedSynchronizer.Node next2 = prev.next;
        next.waitStatus = 1;
        if (next == this.tail && this.compareAndSetTail(next, prev)) {
            compareAndSetNext(prev, next2, null);
        }
        else {
            int waitStatus;
            if (prev != this.head && ((waitStatus = prev.waitStatus) == -1 || (waitStatus <= 0 && compareAndSetWaitStatus(prev, waitStatus, -1))) && prev.thread != null) {
                AbstractQueuedSynchronizer.Node next3 = next.next;
                if (next3 != null && next3.waitStatus <= 0) {
                    compareAndSetNext(prev, next2, next3);
                }
            }
            else {
                this.unparkSuccessor(next);
            }
            next.next = next;
        }
        return;
    }
    
    private static boolean shouldParkAfterFailedAcquire(AbstractQueuedSynchronizer.Node node, AbstractQueuedSynchronizer.Node next) {
        int waitStatus = node.waitStatus;
        if (waitStatus == -1) {
            return true;
        }
        if (waitStatus > 0) {
            do {
                node = (next.prev = node.prev);
            } while (node.waitStatus > 0);
            node.next = next;
        }
        else {
            compareAndSetWaitStatus(node, waitStatus, -1);
        }
        return false;
    }
    
    static void selfInterrupt() {
        Thread.currentThread().interrupt();
    }
    
    private final boolean parkAndCheckInterrupt() {
        LockSupport.park(this);
        return Thread.interrupted();
    }
    
    final boolean acquireQueued(AbstractQueuedSynchronizer.Node head, int n) {
        boolean n2 = true;
        try {
            boolean n3 = false;
            AbstractQueuedSynchronizer.Node predecessor;
            while (true) {
                predecessor = head.predecessor();
                if (predecessor == this.head && this.tryAcquire(n)) {
                    break;
                }
                if (shouldParkAfterFailedAcquire(predecessor, head) && this.parkAndCheckInterrupt()) {
                    n3 = true;
                }
                else {
                    continue;
                }
            }
            this.setHead(head);
            predecessor.next = null;
            n2 = false;
            return n3;
        }
        finally {
            if (!(n2 == 0)) {
                this.cancelAcquire(head);
            }
        }
    }
    
    private void doAcquireInterruptibly(int n) throws InterruptedException {
        AbstractQueuedSynchronizer.Node addWaiter = this.addWaiter(AbstractQueuedSynchronizer.Node.EXCLUSIVE);
        boolean n2 = true;
        while (true) {
            try {
                AbstractQueuedSynchronizer.Node predecessor = addWaiter.predecessor();
                if (predecessor == this.head && this.tryAcquire(n)) {
                    this.setHead(addWaiter);
                    predecessor.next = null;
                    n2 = false;
                    return;
                }
                if (shouldParkAfterFailedAcquire(predecessor, addWaiter) && this.parkAndCheckInterrupt()) {
                    throw new InterruptedException();
                }
                continue;
            }
            finally {
                if (!(n2 == 0)) {
                    this.cancelAcquire(addWaiter);
                }
            }
        }
    }
    
    private boolean doAcquireNanos(int n, long n2) throws InterruptedException {
        if (n2 <= 0L) {
            return false;
        }
        long n3 = System.nanoTime() + n2;
        AbstractQueuedSynchronizer.Node addWaiter = this.addWaiter(AbstractQueuedSynchronizer.Node.EXCLUSIVE);
        boolean n4 = true;
        while (true) {
            try {
                AbstractQueuedSynchronizer.Node predecessor = addWaiter.predecessor();
                if (predecessor == this.head && this.tryAcquire(n)) {
                    this.setHead(addWaiter);
                    predecessor.next = null;
                    n4 = false;
                    return true;
                }
                n2 = n3 - System.nanoTime();
                if (n2 <= 0L) {
                    return false;
                }
                if (shouldParkAfterFailedAcquire(predecessor, addWaiter) && n2 > 1000L) {
                    LockSupport.parkNanos(this, n2);
                }
                if (Thread.interrupted()) {
                    throw new InterruptedException();
                }
                continue;
            }
            finally {
                if (!(n4 == 0)) {
                    this.cancelAcquire(addWaiter);
                }
            }
        }
    }
    
    private void doAcquireShared(int n) {
        AbstractQueuedSynchronizer.Node addWaiter = this.addWaiter(AbstractQueuedSynchronizer.Node.SHARED);
        boolean n2 = true;
        try {
            int n3 = 0;
            AbstractQueuedSynchronizer.Node predecessor;
            int tryAcquireShared;
            while (true) {
                predecessor = addWaiter.predecessor();
                if (predecessor == this.head) {
                    tryAcquireShared = this.tryAcquireShared(n);
                    if (tryAcquireShared >= 0) {
                        break;
                    }
                }
                if (shouldParkAfterFailedAcquire(predecessor, addWaiter) && this.parkAndCheckInterrupt()) {
                    n3 = 1;
                }
                else {
                    continue;
                }
            }
            this.setHeadAndPropagate(addWaiter, tryAcquireShared);
            predecessor.next = null;
            if (n3 != 0) {
                selfInterrupt();
            }
            n2 = false;
            return;
        }
        finally {
            if (!(n2 == 0)) {
                this.cancelAcquire(addWaiter);
            }
        }
    }
    
    private void doAcquireSharedInterruptibly(int n) throws InterruptedException {
        AbstractQueuedSynchronizer.Node addWaiter = this.addWaiter(AbstractQueuedSynchronizer.Node.SHARED);
        boolean n2 = true;
        while (true) {
            try {
                AbstractQueuedSynchronizer.Node predecessor = addWaiter.predecessor();
                if (predecessor == this.head) {
                    int tryAcquireShared = this.tryAcquireShared(n);
                    if (tryAcquireShared >= 0) {
                        this.setHeadAndPropagate(addWaiter, tryAcquireShared);
                        predecessor.next = null;
                        n2 = false;
                        return;
                    }
                }
                if (shouldParkAfterFailedAcquire(predecessor, addWaiter) && this.parkAndCheckInterrupt()) {
                    throw new InterruptedException();
                }
                continue;
            }
            finally {
                if (!(n2 == 0)) {
                    this.cancelAcquire(addWaiter);
                }
            }
        }
    }
    
    private boolean doAcquireSharedNanos(int n, long n2) throws InterruptedException {
        if (n2 <= 0L) {
            return false;
        }
        long n3 = System.nanoTime() + n2;
        AbstractQueuedSynchronizer.Node addWaiter = this.addWaiter(AbstractQueuedSynchronizer.Node.SHARED);
        boolean n4 = true;
        while (true) {
            try {
                AbstractQueuedSynchronizer.Node predecessor = addWaiter.predecessor();
                if (predecessor == this.head) {
                    int tryAcquireShared = this.tryAcquireShared(n);
                    if (tryAcquireShared >= 0) {
                        this.setHeadAndPropagate(addWaiter, tryAcquireShared);
                        predecessor.next = null;
                        n4 = false;
                        return true;
                    }
                }
                n2 = n3 - System.nanoTime();
                if (n2 <= 0L) {
                    return false;
                }
                if (shouldParkAfterFailedAcquire(predecessor, addWaiter) && n2 > 1000L) {
                    LockSupport.parkNanos(this, n2);
                }
                if (Thread.interrupted()) {
                    throw new InterruptedException();
                }
                continue;
            }
            finally {
                if (!(n4 == 0)) {
                    this.cancelAcquire(addWaiter);
                }
            }
        }
    }
    
    protected boolean tryAcquire(int n) {
        throw new UnsupportedOperationException();
    }
    
    protected boolean tryRelease(int n) {
        throw new UnsupportedOperationException();
    }
    
    protected int tryAcquireShared(int n) {
        throw new UnsupportedOperationException();
    }
    
    protected boolean tryReleaseShared(int n) {
        throw new UnsupportedOperationException();
    }
    
    protected boolean isHeldExclusively() {
        throw new UnsupportedOperationException();
    }
    
    public final void acquire(int n) {
        if (!this.tryAcquire(n) && this.acquireQueued(this.addWaiter(AbstractQueuedSynchronizer.Node.EXCLUSIVE), n)) {
            selfInterrupt();
        }
    }
    
    public final void acquireInterruptibly(int n) throws InterruptedException {
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
        if (!this.tryAcquire(n)) {
            this.doAcquireInterruptibly(n);
        }
        return;
    }
    
    public final boolean tryAcquireNanos(int n, long n2) throws InterruptedException {
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
        return this.tryAcquire(n) || this.doAcquireNanos(n, n2);
    }
    
    public final boolean release(int n) {
        if (this.tryRelease(n)) {
            AbstractQueuedSynchronizer.Node head = this.head;
            if (head != null && head.waitStatus != 0) {
                this.unparkSuccessor(head);
            }
            return true;
        }
        else {
            return false;
        }
    }
    
    public final void acquireShared(int n) {
        if (this.tryAcquireShared(n) < 0) {
            this.doAcquireShared(n);
        }
    }
    
    public final void acquireSharedInterruptibly(int n) throws InterruptedException {
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
        if (this.tryAcquireShared(n) < 0) {
            this.doAcquireSharedInterruptibly(n);
        }
        return;
    }
    
    public final boolean tryAcquireSharedNanos(int n, long n2) throws InterruptedException {
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
        return this.tryAcquireShared(n) >= 0 || this.doAcquireSharedNanos(n, n2);
    }
    
    public final boolean releaseShared(int n) {
        if (this.tryReleaseShared(n)) {
            this.doReleaseShared();
            return true;
        }
        return false;
    }
    
    public final boolean hasQueuedThreads() {
        return this.head != this.tail;
    }
    
    public final boolean hasContended() {
        return this.head != null;
    }
    
    public final Thread getFirstQueuedThread() {
        return (this.head == this.tail) ? null : this.fullGetFirstQueuedThread();
    }
    
    private Thread fullGetFirstQueuedThread() {
        AbstractQueuedSynchronizer.Node head;
        AbstractQueuedSynchronizer.Node next;
        Thread thread;
        AbstractQueuedSynchronizer.Node head2;
        AbstractQueuedSynchronizer.Node next2;
        if (((head = this.head) != null && (next = head.next) != null && next.prev == this.head && (thread = next.thread) != null) || ((head2 = this.head) != null && (next2 = head2.next) != null && next2.prev == this.head && (thread = next2.thread) != null)) {
            return thread;
        }
        AbstractQueuedSynchronizer.Node node = this.tail;
        Thread thread2 = null;
        while (node != null && node != this.head) {
            Thread thread3 = node.thread;
            if (thread3 != null) {
                thread2 = thread3;
            }
            node = node.prev;
        }
        return thread2;
    }
    
    public final boolean isQueued(Thread thread) {
        if (thread == null) {
            throw new NullPointerException();
        }
        AbstractQueuedSynchronizer.Node node = this.tail;
        while (node != null) {
            if (node.thread == thread) {
                return true;
            }
            node = node.prev;
        }
        return false;
    }
    
    final boolean apparentlyFirstQueuedIsExclusive() {
        AbstractQueuedSynchronizer.Node head;
        AbstractQueuedSynchronizer.Node next;
        return (head = this.head) != null && (next = head.next) != null && !next.isShared() && next.thread != null;
    }
    
    public final boolean hasQueuedPredecessors() {
        AbstractQueuedSynchronizer.Node tail = this.tail;
        AbstractQueuedSynchronizer.Node head = this.head;
        AbstractQueuedSynchronizer.Node next;
        return head != tail && ((next = head.next) == null || next.thread != Thread.currentThread());
    }
    
    public final int getQueueLength() {
        int n = 0;
        for (AbstractQueuedSynchronizer.Node node = this.tail; node != null; node = node.prev) {
            if (node.thread != null) {
                ++n;
            }
        }
        return n;
    }
    
    public final Collection<Thread> getQueuedThreads() {
        ArrayList<Thread> list = new ArrayList();
        for (AbstractQueuedSynchronizer.Node node = this.tail; node != null; node = node.prev) {
            Thread thread = node.thread;
            if (thread != null) {
                list.add(thread);
            }
        }
        return list;
    }
    
    public final Collection<Thread> getExclusiveQueuedThreads() {
        ArrayList<Thread> list = new ArrayList();
        for (AbstractQueuedSynchronizer.Node node = this.tail; node != null; node = node.prev) {
            if (!node.isShared()) {
                Thread thread = node.thread;
                if (thread != null) {
                    list.add(thread);
                }
            }
        }
        return list;
    }
    
    public final Collection<Thread> getSharedQueuedThreads() {
        ArrayList<Thread> list = new ArrayList();
        for (AbstractQueuedSynchronizer.Node node = this.tail; node != null; node = node.prev) {
            if (node.isShared()) {
                Thread thread = node.thread;
                if (thread != null) {
                    list.add(thread);
                }
            }
        }
        return list;
    }
    
    @Override
    public String toString() {
        return super.toString() + "[State = " + this.getState() + ", " + (this.hasQueuedThreads() ? "non" : "") + "empty queue]";
    }
    
    final boolean isOnSyncQueue(AbstractQueuedSynchronizer.Node node) {
        return node.waitStatus != -2 && node.prev != null && (node.next != null || this.findNodeFromTail(node));
    }
    
    private boolean findNodeFromTail(AbstractQueuedSynchronizer.Node node) {
        AbstractQueuedSynchronizer.Node node2 = this.tail;
        while (node2 != node) {
            if (node2 == null) {
                return false;
            }
            node2 = node2.prev;
        }
        return true;
    }
    
    final boolean transferForSignal(AbstractQueuedSynchronizer.Node node) {
        if (!compareAndSetWaitStatus(node, -2, 0)) {
            return false;
        }
        AbstractQueuedSynchronizer.Node enq = this.enq(node);
        int waitStatus = enq.waitStatus;
        if (waitStatus > 0 || !compareAndSetWaitStatus(enq, waitStatus, -1)) {
            LockSupport.unpark(node.thread);
        }
        return true;
    }
    
    final boolean transferAfterCancelledWait(AbstractQueuedSynchronizer.Node node) {
        if (compareAndSetWaitStatus(node, -2, 0)) {
            this.enq(node);
            return true;
        }
        while (!this.isOnSyncQueue(node)) {
            Thread.yield();
        }
        return false;
    }
    
    final int fullyRelease(AbstractQueuedSynchronizer.Node node) {
        boolean n = true;
        try {
            int state = this.getState();
            if (this.release(state)) {
                n = false;
                return state;
            }
            throw new IllegalMonitorStateException();
        }
        finally {
            if (!(n == 0)) {
                node.waitStatus = 1;
            }
        }
    }
    
    public final boolean owns(AbstractQueuedSynchronizer.ConditionObject conditionObject) {
        return conditionObject.isOwnedBy(this);
    }
    
    public final boolean hasWaiters(AbstractQueuedSynchronizer.ConditionObject conditionObject) {
        if (!this.owns(conditionObject)) {
            throw new IllegalArgumentException("Not owner");
        }
        return conditionObject.hasWaiters();
    }
    
    public final int getWaitQueueLength(AbstractQueuedSynchronizer.ConditionObject conditionObject) {
        if (!this.owns(conditionObject)) {
            throw new IllegalArgumentException("Not owner");
        }
        return conditionObject.getWaitQueueLength();
    }
    
    public final Collection<Thread> getWaitingThreads(AbstractQueuedSynchronizer.ConditionObject conditionObject) {
        if (!this.owns(conditionObject)) {
            throw new IllegalArgumentException("Not owner");
        }
        return conditionObject.getWaitingThreads();
    }
    
    private final boolean compareAndSetHead(AbstractQueuedSynchronizer.Node x) {
        return AbstractQueuedSynchronizer.unsafe.compareAndSwapObject(this, AbstractQueuedSynchronizer.headOffset, null, x);
    }
    
    private final boolean compareAndSetTail(AbstractQueuedSynchronizer.Node expected, AbstractQueuedSynchronizer.Node x) {
        return AbstractQueuedSynchronizer.unsafe.compareAndSwapObject(this, AbstractQueuedSynchronizer.tailOffset, expected, x);
    }
    
    private static final boolean compareAndSetWaitStatus(AbstractQueuedSynchronizer.Node o, int expected, int x) {
        return AbstractQueuedSynchronizer.unsafe.compareAndSwapInt(o, AbstractQueuedSynchronizer.waitStatusOffset, expected, x);
    }
    
    private static final boolean compareAndSetNext(AbstractQueuedSynchronizer.Node o, AbstractQueuedSynchronizer.Node expected, AbstractQueuedSynchronizer.Node x) {
        return AbstractQueuedSynchronizer.unsafe.compareAndSwapObject(o, AbstractQueuedSynchronizer.nextOffset, expected, x);
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        try {
            stateOffset = AbstractQueuedSynchronizer.unsafe.objectFieldOffset(AbstractQueuedSynchronizer.class.getDeclaredField("state"));
            headOffset = AbstractQueuedSynchronizer.unsafe.objectFieldOffset(AbstractQueuedSynchronizer.class.getDeclaredField("head"));
            tailOffset = AbstractQueuedSynchronizer.unsafe.objectFieldOffset(AbstractQueuedSynchronizer.class.getDeclaredField("tail"));
            waitStatusOffset = AbstractQueuedSynchronizer.unsafe.objectFieldOffset(AbstractQueuedSynchronizer.Node.class.getDeclaredField("waitStatus"));
            nextOffset = AbstractQueuedSynchronizer.unsafe.objectFieldOffset(AbstractQueuedSynchronizer.Node.class.getDeclaredField("next"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class Node
    {
        static final AbstractQueuedSynchronizer.Node SHARED;
        static final AbstractQueuedSynchronizer.Node EXCLUSIVE;
        static final int CANCELLED = 1;
        static final int SIGNAL = -1;
        static final int CONDITION = -2;
        static final int PROPAGATE = -3;
        volatile int waitStatus;
        volatile AbstractQueuedSynchronizer.Node prev;
        volatile AbstractQueuedSynchronizer.Node next;
        volatile Thread thread;
        AbstractQueuedSynchronizer.Node nextWaiter;
        
        final boolean isShared() {
            return this.nextWaiter == AbstractQueuedSynchronizer.Node.SHARED;
        }
        
        final AbstractQueuedSynchronizer.Node predecessor() throws NullPointerException {
            AbstractQueuedSynchronizer.Node prev = this.prev;
            if (prev == null) {
                throw new NullPointerException();
            }
            return prev;
        }
        
        Node() {
        }
        
        Node(Thread thread, AbstractQueuedSynchronizer.Node nextWaiter) {
            this.nextWaiter = nextWaiter;
            this.thread = thread;
        }
        
        Node(Thread thread, int waitStatus) {
            this.waitStatus = waitStatus;
            this.thread = thread;
        }
        
        static {
            SHARED = new AbstractQueuedSynchronizer.Node();
            EXCLUSIVE = null;
        }
    }
    
    public class ConditionObject implements Condition, Serializable
    {
        private static final long serialVersionUID = 1173984872572414699L;
        private transient AbstractQueuedSynchronizer.Node firstWaiter;
        private transient AbstractQueuedSynchronizer.Node lastWaiter;
        private static final int REINTERRUPT = 1;
        private static final int THROW_IE = -1;
        
        private AbstractQueuedSynchronizer.Node addConditionWaiter() {
            AbstractQueuedSynchronizer.Node node = this.lastWaiter;
            if (node != null && node.waitStatus != -2) {
                this.unlinkCancelledWaiters();
                node = this.lastWaiter;
            }
            AbstractQueuedSynchronizer.Node lastWaiter = new AbstractQueuedSynchronizer.Node(Thread.currentThread(), -2);
            if (node == null) {
                this.firstWaiter = lastWaiter;
            }
            else {
                node.nextWaiter = lastWaiter;
            }
            return this.lastWaiter = lastWaiter;
        }
        
        private void doSignal(AbstractQueuedSynchronizer.Node firstWaiter) {
            do {
                if ((this.firstWaiter = firstWaiter.nextWaiter) == null) {
                    this.lastWaiter = null;
                }
                firstWaiter.nextWaiter = null;
            } while (!AbstractQueuedSynchronizer.this.transferForSignal(firstWaiter) && (firstWaiter = this.firstWaiter) != null);
        }
        
        private void doSignalAll(AbstractQueuedSynchronizer.Node node) {
            AbstractQueuedSynchronizer.Node node2 = null;
            this.firstWaiter = node2;
            this.lastWaiter = node2;
            do {
                AbstractQueuedSynchronizer.Node nextWaiter = node.nextWaiter;
                node.nextWaiter = null;
                AbstractQueuedSynchronizer.this.transferForSignal(node);
                node = nextWaiter;
            } while (node != null);
        }
        
        private void unlinkCancelledWaiters() {
            AbstractQueuedSynchronizer.Node firstWaiter = this.firstWaiter;
            AbstractQueuedSynchronizer.Node lastWaiter = null;
            while (firstWaiter != null) {
                AbstractQueuedSynchronizer.Node nextWaiter = firstWaiter.nextWaiter;
                if (firstWaiter.waitStatus != -2) {
                    firstWaiter.nextWaiter = null;
                    if (lastWaiter == null) {
                        this.firstWaiter = nextWaiter;
                    }
                    else {
                        lastWaiter.nextWaiter = nextWaiter;
                    }
                    if (nextWaiter == null) {
                        this.lastWaiter = lastWaiter;
                    }
                }
                else {
                    lastWaiter = firstWaiter;
                }
                firstWaiter = nextWaiter;
            }
        }
        
        @Override
        public final void signal() {
            if (!AbstractQueuedSynchronizer.this.isHeldExclusively()) {
                throw new IllegalMonitorStateException();
            }
            AbstractQueuedSynchronizer.Node firstWaiter = this.firstWaiter;
            if (firstWaiter != null) {
                this.doSignal(firstWaiter);
            }
            return;
        }
        
        @Override
        public final void signalAll() {
            if (!AbstractQueuedSynchronizer.this.isHeldExclusively()) {
                throw new IllegalMonitorStateException();
            }
            AbstractQueuedSynchronizer.Node firstWaiter = this.firstWaiter;
            if (firstWaiter != null) {
                this.doSignalAll(firstWaiter);
            }
            return;
        }
        
        @Override
        public final void awaitUninterruptibly() {
            AbstractQueuedSynchronizer.Node addConditionWaiter = this.addConditionWaiter();
            int fullyRelease = AbstractQueuedSynchronizer.this.fullyRelease(addConditionWaiter);
            boolean n = false;
            while (!AbstractQueuedSynchronizer.this.isOnSyncQueue(addConditionWaiter)) {
                LockSupport.park(this);
                if (!Thread.interrupted()) {
                    continue;
                }
                n = true;
            }
            if (AbstractQueuedSynchronizer.this.acquireQueued(addConditionWaiter, fullyRelease) || n) {
                AbstractQueuedSynchronizer.selfInterrupt();
            }
        }
        
        private int checkInterruptWhileWaiting(AbstractQueuedSynchronizer.Node node) {
            return Thread.interrupted() ? (AbstractQueuedSynchronizer.this.transferAfterCancelledWait(node) ? -1 : 1) : 0;
        }
        
        private void reportInterruptAfterWait(int n) throws InterruptedException {
            if (n == -1) {
                throw new InterruptedException();
            }
            if (n == 1) {
                AbstractQueuedSynchronizer.selfInterrupt();
            }
            return;
        }
        
        @Override
        public final void await() throws InterruptedException {
            if (Thread.interrupted()) {
                throw new InterruptedException();
            }
            AbstractQueuedSynchronizer.Node addConditionWaiter = this.addConditionWaiter();
            int fullyRelease = AbstractQueuedSynchronizer.this.fullyRelease(addConditionWaiter);
            int checkInterruptWhileWaiting = 0;
            while (!AbstractQueuedSynchronizer.this.isOnSyncQueue(addConditionWaiter)) {
                LockSupport.park(this);
                if ((checkInterruptWhileWaiting = this.checkInterruptWhileWaiting(addConditionWaiter)) != 0) {
                    break;
                }
                continue;
            }
            if (AbstractQueuedSynchronizer.this.acquireQueued(addConditionWaiter, fullyRelease) && checkInterruptWhileWaiting != -1) {
                checkInterruptWhileWaiting = 1;
            }
            if (addConditionWaiter.nextWaiter != null) {
                this.unlinkCancelledWaiters();
            }
            if (checkInterruptWhileWaiting != 0) {
                this.reportInterruptAfterWait(checkInterruptWhileWaiting);
            }
            return;
        }
        
        @Override
        public final long awaitNanos(long n) throws InterruptedException {
            if (Thread.interrupted()) {
                throw new InterruptedException();
            }
            AbstractQueuedSynchronizer.Node addConditionWaiter = this.addConditionWaiter();
            int fullyRelease = AbstractQueuedSynchronizer.this.fullyRelease(addConditionWaiter);
            long n2 = System.nanoTime() + n;
            int checkInterruptWhileWaiting = 0;
            while (!AbstractQueuedSynchronizer.this.isOnSyncQueue(addConditionWaiter)) {
                if (n <= 0L) {
                    AbstractQueuedSynchronizer.this.transferAfterCancelledWait(addConditionWaiter);
                    break;
                }
                if (n >= 1000L) {
                    LockSupport.parkNanos(this, n);
                }
                if ((checkInterruptWhileWaiting = this.checkInterruptWhileWaiting(addConditionWaiter)) != 0) {
                    break;
                }
                n = n2 - System.nanoTime();
            }
            if (AbstractQueuedSynchronizer.this.acquireQueued(addConditionWaiter, fullyRelease) && checkInterruptWhileWaiting != -1) {
                checkInterruptWhileWaiting = 1;
            }
            if (addConditionWaiter.nextWaiter != null) {
                this.unlinkCancelledWaiters();
            }
            if (checkInterruptWhileWaiting != 0) {
                this.reportInterruptAfterWait(checkInterruptWhileWaiting);
            }
            return n2 - System.nanoTime();
        }
        
        @Override
        public final boolean awaitUntil(Date date) throws InterruptedException {
            long time = date.getTime();
            if (Thread.interrupted()) {
                throw new InterruptedException();
            }
            AbstractQueuedSynchronizer.Node addConditionWaiter = this.addConditionWaiter();
            int fullyRelease = AbstractQueuedSynchronizer.this.fullyRelease(addConditionWaiter);
            boolean transferAfterCancelledWait = false;
            int checkInterruptWhileWaiting = 0;
            while (!AbstractQueuedSynchronizer.this.isOnSyncQueue(addConditionWaiter)) {
                if (System.currentTimeMillis() > time) {
                    transferAfterCancelledWait = AbstractQueuedSynchronizer.this.transferAfterCancelledWait(addConditionWaiter);
                    break;
                }
                LockSupport.parkUntil(this, time);
                if ((checkInterruptWhileWaiting = this.checkInterruptWhileWaiting(addConditionWaiter)) != 0) {
                    break;
                }
                continue;
            }
            if (AbstractQueuedSynchronizer.this.acquireQueued(addConditionWaiter, fullyRelease) && checkInterruptWhileWaiting != -1) {
                checkInterruptWhileWaiting = 1;
            }
            if (addConditionWaiter.nextWaiter != null) {
                this.unlinkCancelledWaiters();
            }
            if (checkInterruptWhileWaiting != 0) {
                this.reportInterruptAfterWait(checkInterruptWhileWaiting);
            }
            return transferAfterCancelledWait == 0;
        }
        
        @Override
        public final boolean await(long duration, TimeUnit timeUnit) throws InterruptedException {
            long nanos = timeUnit.toNanos(duration);
            if (Thread.interrupted()) {
                throw new InterruptedException();
            }
            AbstractQueuedSynchronizer.Node addConditionWaiter = this.addConditionWaiter();
            int fullyRelease = AbstractQueuedSynchronizer.this.fullyRelease(addConditionWaiter);
            long n = System.nanoTime() + nanos;
            boolean transferAfterCancelledWait = false;
            int checkInterruptWhileWaiting = 0;
            while (!AbstractQueuedSynchronizer.this.isOnSyncQueue(addConditionWaiter)) {
                if (nanos <= 0L) {
                    transferAfterCancelledWait = AbstractQueuedSynchronizer.this.transferAfterCancelledWait(addConditionWaiter);
                    break;
                }
                if (nanos >= 1000L) {
                    LockSupport.parkNanos(this, nanos);
                }
                if ((checkInterruptWhileWaiting = this.checkInterruptWhileWaiting(addConditionWaiter)) != 0) {
                    break;
                }
                nanos = n - System.nanoTime();
            }
            if (AbstractQueuedSynchronizer.this.acquireQueued(addConditionWaiter, fullyRelease) && checkInterruptWhileWaiting != -1) {
                checkInterruptWhileWaiting = 1;
            }
            if (addConditionWaiter.nextWaiter != null) {
                this.unlinkCancelledWaiters();
            }
            if (checkInterruptWhileWaiting != 0) {
                this.reportInterruptAfterWait(checkInterruptWhileWaiting);
            }
            return transferAfterCancelledWait == 0;
        }
        
        final boolean isOwnedBy(AbstractQueuedSynchronizer abstractQueuedSynchronizer) {
            return abstractQueuedSynchronizer == AbstractQueuedSynchronizer.this;
        }
        
        protected final boolean hasWaiters() {
            if (!AbstractQueuedSynchronizer.this.isHeldExclusively()) {
                throw new IllegalMonitorStateException();
            }
            AbstractQueuedSynchronizer.Node node = this.firstWaiter;
            while (node != null) {
                if (node.waitStatus == -2) {
                    return true;
                }
                node = node.nextWaiter;
            }
            return false;
        }
        
        protected final int getWaitQueueLength() {
            if (!AbstractQueuedSynchronizer.this.isHeldExclusively()) {
                throw new IllegalMonitorStateException();
            }
            int n = 0;
            for (AbstractQueuedSynchronizer.Node node = this.firstWaiter; node != null; node = node.nextWaiter) {
                if (node.waitStatus == -2) {
                    ++n;
                }
            }
            return n;
        }
        
        protected final Collection<Thread> getWaitingThreads() {
            if (!AbstractQueuedSynchronizer.this.isHeldExclusively()) {
                throw new IllegalMonitorStateException();
            }
            ArrayList<Thread> list = new ArrayList();
            for (AbstractQueuedSynchronizer.Node node = this.firstWaiter; node != null; node = node.nextWaiter) {
                if (node.waitStatus == -2) {
                    Thread thread = node.thread;
                    if (thread != null) {
                        list.add(thread);
                    }
                }
            }
            return list;
        }
    }
}
